// --- [adxl.c]----------------------------------------------------------------
//
//	tab = 3
//
// Der ADXL wird hier als Alarmgeber eingesetzt
// Beim Auslsen des Alarms zieht der ADXL mit seinem INT1 den INT0 des Controllers
// auf gnd und lst damit einen INT0 (low-level) Interrupt aus.
// Gleiches ereignet sich am Ende des Alarmereignisses:
// der ADXL zieht mit seinem INT2 den INT1 des Controllers auf gnd und lst
// damit einen INT1-Interrupt aus. 
// 
// Der Controller hat die Aufgabe:
// - beim Programmstart den ADXL zu initialisieren
// - nach einem Interrupt das Register INT_Source auslesen, damit die Flags 
//   gelscht werden.
// - ein akustisches Alarmzeichen geben.
// - bei Bedarf knnen zustzliche Schaltvorgnge eingebaut werden
//
// Der Controller erzeugt ein akustisches Signal
// und schaltet ggf. mit Verzgerung einen Schalter ab.
// Die Funktionalitt des Programmes ist daher stark reduziert. 

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>
#include <util/delay.h>
#include "global.h"

#include "adxl.h"
#include "rs232_int.h"
#include "soft_spi_master.h"

int16_t s[3];										// fr ein Datensatz Messwerte
int16_t delta[3] = {0,0,0};					// ggf. Korrekturwerte dazu
char msg[10];										// zur Umwandlung von Zahlen in Zeichenfolge
														// Flags fr Interrupts, Counter fr Alarmsummer
volatile uint8_t flag1 = 0, flag2 = 0; 
volatile uint8_t timer0_cnt = 0;
volatile uint8_t timer2_cnt;

uint8_t ADXL_write_register(uint8_t reg, uint8_t value);
uint8_t ADXL_read_register(uint8_t reg);
void    ADXL_read_data(void);
void    ADXL_clear_buffer(void);

void rs232_print_ADXL_DATA(void);
void wait_10ms(uint8_t t);
void alarm_on(void);
void alarm_off(void);
void start_timer2(void);
void stop_timer2(void);

// ----------------------------------------------------------------------------
// schreibt einen Wert in genau ein Register des ADXL
// prft, ob derselbe Wert auch wieder ausgelesen werden kann
uint8_t ADXL_write_register(uint8_t reg, uint8_t value)
{
uint8_t a[2];

reg &= 0x3F;										// obere beiden Bits maskieren

#ifdef DEBUG
put_s("0x");
put_s(utoa(reg, msg, 16));
put_s(" -> ");
put_s(utoa(value, msg,16));
#endif

// das Write-Bit lschen und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 32 ist !) addieren
//a[0] = 0x01000000 + reg;
a[0] = 0x40 + reg;
a[1] = value;										// der Wert des Register

ADXL_SPI_write_array(a, 2);					// das Array schreiben

a[0] = 0xC0 + reg;								// Zum Test das geschriebene Byte 
ADXL_SPI_write_array(a, 2);					// neu einlesen und vergleichen

#ifdef DEBUG
if (a[1] == value) put_s(" ok"); else put_s(" err");
put_c(13);
#endif

if (a[1] == value) return(NO_ERROR); else return(ERROR);

}

// ----------------------------------------------------------------------------
// liest den Inhalt genau eines Registers des ADXL
uint8_t ADXL_read_register(uint8_t reg)
{
uint8_t a[2];
// das Write-Bit und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 64 sein muss !) addieren

a[0] = 0xC0 + (reg & 0x3F);
ADXL_SPI_write_array(a, 2);

return(a[1]);										// das gelesene Byte steht in a[1]

}

// ----------------------------------------------------------------------------
// liest einen Datensatz von 3 * 2 Byte mit den x, y, z-Werten
// wandelt die Byte-Paare jeweils in int16-Werte um
// berechnet die Quadrate der einzelnen Vektorkomponenten und
// ermittelt die Lnge des Vektors als Summe der Quadrate der Komponenten
// und zieht davon das Quadrat des Ruhewertes 1g ab.
void ADXL_read_data(void)
{
uint8_t i,j;
uint8_t a[7];

// das Write-Bit und das Multiple_Byte Bit setzen
// und die Registeradresse (die < 64 sein muss !) addieren
a[0] = (0xC0 + ADXL_DATA);
ADXL_SPI_write_array(a, 7);					// 6 Byte ab ADXL_DATA holen

j = 0;

for (i = 1; i < 7; i = i + 2)
	{
	s[j]  = a[i + 1] * 256;						// das High-Byte schieben
	s[j] += a[i];									// das Low-Byte addieren
	s[j] -= delta[j];								// ggf. Korrekturwert bercksichtigen
	j++;
	}
}

// ----------------------------------------------------------------------------
// Initialisiert den ADXL345 fr Interrupt-Modus
// Daten werden nicht ausgelesen
// Activity- und Inactivity-Interrupt werden genutzt.
// Bei Bewegung des Sensors schaltet der Interrupt
// Activity einen Alarm (Summer) an, 
// nach Ablauf der Zeit TIME_INACT schaltet der Interrupt Inactivity ihn wieder ab
uint8_t ADXL_init(void)
{
uint8_t err = 0;

// der Chip meldet sich mit einer festen ID,  wenn nicht, dann Fehler
if (!(ADXL_read_register(ADXL_DEVID) == ADXL_ID)) err = 128;


//err += ADXL_write_register(ADXL_TRESH_TAP,   250);		// 20 * 62.5mg
//err += ADXL_write_register(ADXL_TAP_AXES, 0x00);
err += ADXL_write_register(ADXL_DATA_FORMAT, DATA_INT_INVERT | DATA_2G);

err += ADXL_write_register(ADXL_THRESH_ACT,   5);		// 5 * 62.5mg
err += ADXL_write_register(ADXL_THRESH_INACT, 5);	   // 5 * 62.5mg
err += ADXL_write_register(ADXL_TIME_INACT,   10);		// 10 * 1sec
err += ADXL_write_register(ADXL_ACT_INACT_CTL, 0xFF); //  AC-coupled, alle Achsen aktiv
err += ADXL_write_register(ADXL_INT_MAP,       0x08);	// Activity = INT0, Inactivity = INT1
err += ADXL_write_register(ADXL_INT_ENABLE,    0x18);	// Activity, Inactivity enablen

err += ADXL_write_register(ADXL_BW_RATE, BW_RATE_LOW_POWER | BW_RATE_12_5);
err += ADXL_write_register(ADXL_FIFO_CTL,  FIFO_CTL_FIFO_BYPASS); // keine Daten puffern
err += ADXL_write_register(ADXL_POWER_CTL, POWER_CTL_LINK | POWER_CTL_MEASURE);

return(err);
}

// ----------------------------------------------------------------------------
// Liest alle 32 Datenstze ein, damit alle Interruptflags gelscht werden fr die
// Ereignisse: dataready, watermark, overflow.
// Zum Lschen der anderen Interrupts muss das Register INT_SOURCE gelesen werden !

void ADXL_clear_buffer(void)
{
uint8_t i = 32;
while(i--)	ADXL_read_data();

}

// ----------------------------------------------------------------------------
// gibt zum Debuggen alle Register des ADXL345 mit Adresse und aktuellem Wert aus
#ifdef DEBUG
void ADXL_print_registers(void)
{
uint8_t i;
for (i = ADXL_DEVID; i < 0x3A; i++)
	{
	put_s("0x");
	put_s(itoa(i, msg, 16));
	put_s(" = 0x");
	put_s(itoa(ADXL_read_register(i), msg, 16));
	put_c(13);
	
	if (i == ADXL_DEVID) i = (0x1d - 1);	// die Lcke zwischen ADXL_DEVID und 0x1D berspringen
	}
}
#endif

// ----------------------------------------------------------------------------
int main( void )
{
uint8_t i; 

// nicht bentigte Module abschalten um Strom zu sparen
PRR = (1<<PRTWI | 0<<PRTIM2 | 0<<PRTIM0 | 1<<PRTIM1 | 0<<PRSPI | 0<<PRUSART0 | 1<<PRADC);
ACSR = (1<<ACD);                       	// Analog Comparator abschalten

SPI_master_init();

sei();												// Interrupt fr (TWI/RS232) INT0/1 notwendig !!

rs232_init();										// auf 9600 Baud

/*
put_s("Hello world");							// zeigt, dass die serielle Schnittstelle lebt
put_c(13);
wait_10ms(50);
*/

#ifdef DEBUG
put_s("ADXL ID: ");
put_s(utoa(ADXL_read_register(0), msg, 16));
put_c(13);


// Die Registerwerte vor der Initialisierung ausgeben
put_s("Register vor Init:");
put_c(13);
ADXL_print_registers();
put_c(13);
put_s("writing registers:");
put_c(13);
#endif

i = ADXL_init();									// hier werden die Register beschrieben
														// dabei wird protokolliert
put_c(13);
put_s("ADXL ");
if (i)												// falls i > 0, dann gab es Fehler
	{
	put_s("error: ");
	put_s(utoa(i, msg, 10));
	}
else
	{
	put_s("ok");
	}
put_c(13);

/*
// fr eine "echte" Anwendung kann hier eine Wartezeit bis zum Scharfschalten vorgesehen werden
i = 120;
while(i--) wait_10ms(100);						// 120 Sekunden warten
*/

#ifdef DEBUG
put_s("Register nach Init");					// die Registerwerte nach der Initialisierung ausgeben
put_c(13);
ADXL_print_registers();
#endif


//INT_PORT |= 1<<PIN_INT0 | 1<<PIN_INT1;		// Pullups setzen -> erhht die Ruhestromaufnahme deutlich !!
ADXL_clear_buffer();								// damit werden alle aufgelaufenen Interrupts freigegeben
ADXL_read_register(ADXL_INT_SOURCE); 		// und hier alle Flags gelscht werden
														// INT0/INT1 am uc fr INT1/INT2 am ADXL konfigurieren

EICRA = (0<< ISC11 | 0<<ISC10 | 0<< ISC01 | 0<<ISC00); // low level
EIMSK = (1<< INT1  | 1<<INT0);				// INT0 und INT1 aktivieren

while(!(UCSR0A & (1<<UDRE0)))	wait_10ms(10);// abwarten, bis letzte RS232-bertragung abgeschlossen ist				

timer2_cnt = 0;									// der Controller geht gleich in den Sleep-Modus

while(1)
	{

	if (flag2)										// wenn der Alarm beendet wurde
		{
		flag2= 0;
		i = ADXL_read_register(ADXL_INT_SOURCE); // hier werden die Flags gelscht
		EIMSK = (1<< INT1  | 1<<INT0);		// jetzt die Interrupts wieder aktivieren
		
		put_s("INT2: ") ;
		put_s(itoa(i, msg, 16));				// die Flags ausgeben
		put_c(13);
		
		ADXL_read_data();							// einen Datensatz einlesen
		rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen
			
		alarm_off();
														// Entweder:
		//start_timer2();							// Timer2 einschalten, der zhlt einen Counter runter	
														// hier kann noch eine zustzliche Aktion geschaltet werden
		timer2_cnt = 0;							// oder: den Counter auf 0 setzen, dann wird
		}												// direkt ins Sleep verzweigt.

	if (flag1)										// wenn ein Alarm erkannt wurde
		{
		flag1 = 0;
		i = ADXL_read_register(ADXL_INT_SOURCE); // hier werden die Flags gelscht
		EIMSK = (1<< INT1  | 1<<INT0);		// die Interrupts wieder aktivieren		

		put_s("INT1: ") ;
		put_s(itoa(i, msg, 16));				// Flags ausgeben
		put_c(13);
		
		ADXL_read_data();							// einen Datensatz einlesen
		rs232_print_ADXL_DATA();				// zur Ausgabe auf rs232 weiterreichen

		stop_timer2();								// den timer abschalten - falls er noch luft			
		alarm_on();

		}

	if (!(timer2_cnt))							// nach Ablauf der zustzlichen Verzgerung
		{												// in den Sleep-Modus gehen
														// vorher ggf. etwas (ab-) schalten 
		timer2_cnt = TIMER2_CNT;
		stop_timer2();								// den timer abschalten
	
		put_s("sleeping ...") ;
		put_c(13);	
														// schalte ggf. noch etwas
		
		wait_10ms(1);								// abwarten, bis letzte RS232-bertragung abgeschlossen ist	

		SPI_master_clear_pins();				// Potentialunterschiede zwischen ADXL und SPI neutralisieren	
														// fhrt zu deutlicher Ruhestromreduktion 
														
		EIFR = 0;									// falls Flags gesetzt sind

		// den Sleep-Modus aktivieren und auf einen low-level-Interrupt warten
		SMCR =  (0<<SM0  | 1<<SM1 | 0<<SM0 | 1<<SE); // Power-Down Modus, Sleep enable

		asm volatile("sleep");					// hier wird auf den nchsten INT gewartet

		SMCR = 0;									// sleep disablen
		SPI_master_init();						// den SPI-PORT wieder aktivieren
		}
	}
}


// ----------------------------------------------------------------------------
// Setzt die notwendigen TimerRegister des mc fr einen akustischen Alarm
// der Summer ist an PD6 (OC0A) angeschlossen 
void alarm_on(void)
{

TCCR0A = (1<<COM0A0 | 1<<COM0B0 | 1<< WGM01);  // toogle on Compare Match, CTC-Modus
TCCR0B = (1<<CS01);								// Vorteiler 8, Takt 3.68 MHz CompareA 64 -> 3600 Hz
OCR0A = TONE_LOW;
DDRD |= (1<<PD6);									// hier ist der Summer angeschlossen !

TIMSK0 = 1<<OCIE0A;								// Compare-Interrupt fr Tonhhenwechsel einspannen
timer0_cnt = 0;
}

// ----------------------------------------------------------------------------
// lscht den Alarm
void alarm_off(void)
{
TCCR0B = 0;											// Timer stoppen
DDRD &=(~(1<<PD6 ));								// Ausgangspin als Input
}

// ----------------------------------------------------------------------------
// jeweils nur ein Flag incrementieren (von 0 auf 1)
// und die low-level Interrupts deaktivieren (bis das Flag im ADXL gelscht ist)
ISR (INT0_vect)
{
flag1++;
EIMSK = 0;											// INT0/1 aktivieren
}

// ----------------------------------------------------------------------------
// jeweils nur ein Flag incrementieren (von 0 auf 1)
// und die low-level Interrupts deaktivieren (bis das Flag im ADXL gelscht ist)
ISR (INT1_vect)
{
flag2++;
EIMSK = 0;											// INT0/1 aktivieren
}

// ----------------------------------------------------------------------------
// nutzt den TIMER0_COMP-Interrupt als Zeitglied, um einen Wechsel
// der Tonfrequenz zu erreichen
ISR (TIMER0_COMPA_vect)
{
register uint8_t t;

/*
t = timer0_cnt;									// jeweils nach berlauf auf 0
t++;													// wird die Tonhhe gendert
														// hier: Wechsel zwischen zwei Tnen
if (!(t))											// Der Counter steht auf 0
	{
	if (OCR0A == TONE_HIGH) OCR0A = TONE_LOW; else OCR0A = TONE_HIGH;
	}
timer0_cnt = t;
*/

t = timer0_cnt;									// jeweils nach berlauf auf 0
t++;													// wird die Tonhhe gendert
														// hier: Rampe von Low nach High
if (!(t))
	{
	OCR0A--;											// der Ton wird immer hher
	if (OCR0A < TONE_HIGH) OCR0A = TONE_LOW; // bis zum oberen Grenzwert
	}
timer0_cnt = t;
}

// ----------------------------------------------------------------------------
// wird zur Zeitzhlung (Wartezeit) genutzt
// zhlt die Variable timer2_cnt herab, 
// Verzgerung je Wert bei 3.686400Hz ca. 1 Sekunde
ISR (TIMER2_OVF_vect)
{
static uint8_t i = 10;
i--;
if(i == 0)
	{
	if (timer2_cnt) timer2_cnt--;
	PINB |= 1<<PB0;
	i = 10;
	}
}

//-----------------------------------------------------------------------------
// Startet den Timer 2
void start_timer2(void)
{
timer2_cnt = TIMER2_CNT;
TCNT2 = 0;
//TCCR2A = (1<<WGM21 | 0<<WGM20);			// Mode 2: CTC 
TCCR2B = (1<<CS22 | 1<<CS21 | 1<<CS20);	// Teiler 1024
//OCR2A  = 255;									// 1 Sekunde
TIMSK2 = (1<<TOIE2);
}


// ----------------------------------------------------------------------------
// Stoppt den Timer 2
void stop_timer2(void)
{
TIMSK2 = 0;
TCCR2B = 0;
}

// ----------------------------------------------------------------------------
// wartet jeweils 10ms
void wait_10ms(uint8_t t)
{
while (t--) _delay_ms(10);

}

// ----------------------------------------------------------------------------
// gibt die in s[] zwischengespeicherten Einzelwerte via RS232 aus
void rs232_print_ADXL_DATA(void)
{
uint8_t i;

for (i = 0; i < 3; i++)
	{
	put_s(ltoa(s[i], msg, 10));
	put_c(',');
	}
put_c(13);

}

// --- [eof] ------------------------------------------------------------------